/***
*		DataFilt.c -- b&p midi data filter 
*		 1993,1994,1995 Another Option
***/

/*
============================================================================
|    DATE   | WHAT
============================================================================
	12 Dec 91 | created
	15 Dec 91 | gave up on vanishing fat prop gadgets; they don't move!
	17 Dec 91 | new ui design & configurable controller numbers
	 4 Jan 92 | please remember to point events at the next tool
	20 Apr 92 | yet another new ui to be consistent with other ao tools
	21 Apr 92 | yet another new ui to be really consistent with other ao tools
------------- really version 1.0 ---------------------------
------------- and here starts version 1.1 ------------------
	 4 Jan 93 | allow no data bytes selection to deal with an entire class of
	          | messages without regard to their data; improved sanity checks
				 | on hi/lo/fold values; now a bandpass or notch filter
   18 Jan 93 | massaged for sas/c v6.0
   14 Feb 93 | massaged for righteous bplib.lib
	15 Nov 93 | new window title
	11 Nov 94 | twiddling; add snapshot & sysex
	14 Aug 94 | changed bandpass limit testing to </> limits instead of <=/>=
	29 Nov 94 | no more data byte 1 testing on controller filtering (stupid)
	26 Mar 95 | don't automatically select byte2 on controller filtering (bunghole)
	26 Aug 95 | don't mess with NoteOffs in doevent() anymore (dipshit)
	11 Nov 95 | really don't mess with them
	31 Mar 99 | uh, you didn't set event->tool in processlist(), dickhead! added
	          | appropriate __stdargs. but the real problem is that EVENT_PADEDIT
				 | wasn't seen in doevent(). odd, but patched up to work...
*/

#define DBUG 0

#define THIS_TOOL_NAME "DAFC"

#include <exec/types.h>
#include <libraries/dos.h>
#include <exec/memory.h>
#include <string.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <stdio.h>

#include "bars.h"
#include "edit.h"
#include "bplib.h"
#include "barsmacros.h"
//#include "datafiltw4.h"
#include "datafiltw5.h"
#include "/icons/filtup.h"
#include "/icons/filtdown.h"
#include "snapshotgadget.h"
#include "proto/datafilt_protos.h"

#if DBUG
 #include "debug.h"
#endif

static UBYTE const versiontag[] = "$VER: MIDIEventFilter 40.01 (03.31.99)";

struct FiltTool {
	struct Tool tool;
	UBYTE action;  /* bits for what to do */
	UBYTE event;   /* what event to filter */
	UBYTE control; /* status byte for channel messages */
	UBYTE controller; /* controller number */
	UBYTE data;    /* bits for which bytes to do it to */
	UBYTE low;     /* low end of range */
	UBYTE high;    /* high end of range */
	UBYTE fold;    /* increment/decrement for folding */
	UBYTE fix;     /* value for fixing data */
};

#define TOOL_SIZE sizeof(struct FiltTool) /* for ToolMaster */

/* bits for tool->action */
#define ACT_FOLD    1
#define ACT_FIX     2
#define ACT_BRANCH  4
#define ACT_DISCARD 8
/* set when hi<lo: we're a notch filter */
#define ACT_NOTCH  128
/* respond to sysex */
#define REALTIME_SYSEX 64
#define DISCARD_SYSEX 32

/* mask the real action bits */
#define ACTION_MASK 0x0F
/* mask to preserve notch and sysex settings */
#define ACTION_HIBITS 0xE0

/* bits for tool->data */
#define DATA_BYTE1 1
#define DATA_BYTE2 2

/* defaults for inittool */
#define DEFAULT_ACTION (ACT_FOLD | DISCARD_SYSEX)
#define DEFAULT_EVENT 0
#define DEFAULT_DATA DATA_BYTE1
#define DEFAULT_CONTROLLER 0 //MIDI_ModWheel


extern struct Functions * far functions;


/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void inittool(tool)

struct FiltTool *tool;

{
	tool->low = 0;
	tool->high = 127;
	tool->fold = 12;
	tool->fix = 64;
	tool->action = DEFAULT_ACTION;
	tool->data = DEFAULT_DATA;
	tool->event = DEFAULT_EVENT;
	tool->control = MIDI_NOTEON;
	tool->controller = DEFAULT_CONTROLLER;
	tool->tool.touched |= TOUCH_INIT;
}

#define tw tool->tool.window

/*==========================================================================
* refreshgadgets
*
*  INPUT: pointer to tool, refresh flag
* RETURN: none
*
* EFFECT: redraws gadgetry -- CALL ONLY IF THE WINDOW IS OPEN AND READY
*=========================================================================*/

static void __regargs refreshgadgets(struct FiltTool *tool, UBYTE refresh)

{
	if (refresh & 1) {
		if (tool->event == 1) tool->control=MIDI_CCHANGE;
		else if (tool->event == 2) tool->control=MIDI_PCHANGE;
		else if (tool->event == 3) tool->control=MIDI_MTOUCH;
		else if (tool->event == 4) tool->control=MIDI_PBEND;
		else tool->control=MIDI_NOTEON; /* ==0 or a weird value */
		functions->ModifyEmbossedProp(tw,EVENT,tool->event,0,0,0,0,0);
		functions->DrawEmbossedProp(tw,EVENT);
		EnableFatProp(tw,CONTROLLER,CONTROLLERL,CONTROLLERR,(tool->event==1));
		functions->ModifyEmbossedProp(tw,CONTROLLER,tool->controller,0,0,0,0,0);
		functions->DrawEmbossedProp(tw,CONTROLLER);
		if ((tool->event == 2) || (tool->event==3)) { /* one-byte messages */
			tool->data &= ~DATA_BYTE2;
			functions->EnableEmbossed(tw,BYTE2,0);
//			refresh |= 2;
		}
		else functions->EnableEmbossed(tw,BYTE2,1);
		if (tool->event == 1) { /* controller number defines byte1 */
			tool->data = (tool->data & ~DATA_BYTE1) /*| DATA_BYTE2*/ ; /* 26Mar95: don't select byte2 automatically*/
			functions->EnableEmbossed(tw,BYTE1,0);
//			refresh |= 2;
		}
		else functions->EnableEmbossed(tw,BYTE1,1);
		if (tool->event >= 1 && tool->event < 4) refresh |= 2;
	}
	if (refresh & 2) {
		functions->SelectEmbossed(tw,BYTE1,tool->data & DATA_BYTE1);
		functions->SelectEmbossed(tw,BYTE2,tool->data & DATA_BYTE2);
	}
	if (refresh & 4) {
		functions->SelectEmbossed(tw,FOLD,tool->action & ACT_FOLD);
		functions->SelectEmbossed(tw,FIX,tool->action & ACT_FIX);
		functions->SelectEmbossed(tw,BRANCH,tool->action & ACT_BRANCH);
		functions->SelectEmbossed(tw,DISCARD,tool->action & ACT_DISCARD);
	}
	if (refresh & 8) {
		functions->ModifyEmbossedProp(tw,LOW,tool->low,0,0,0,0,0);
		functions->ModifyEmbossedProp(tw,HIGH,tool->high,0,0,0,0,0);
		functions->DrawEmbossedProp(tw,LOW);
		functions->DrawEmbossedProp(tw,HIGH);
	}
	if (refresh & 16) {
		if (tool->action & ACT_FOLD) { /* if enabled, make sure value is cool */
			if (!(tool->action & ACT_NOTCH)) {
				if (tool->fold > tool->high - tool->low) {
					tool->fold = tool->high-tool->low;
				}
			}
			else { /* we be a notch filter */
				if (tool->fold > tool->high + (127-tool->low)) {
					tool->fold = tool->high + (127-tool->low);
				}
			}
			functions->ModifyEmbossedProp(tw,FOLDVAL,tool->fold,0,0,0,0,0);
		}
		else if (tool->action & ACT_FIX) {
			functions->ModifyEmbossedProp(tw,FIXVAL,tool->fix,0,0,0,0,0);
		}
		EnableFatProp(tw,FOLDVAL,FOLDVALL,FOLDVALR,(tool->action & ACT_FOLD));
		EnableFatProp(tw,FIXVAL,FIXVALL,FIXVALR,(tool->action & ACT_FIX));
		functions->DrawEmbossedProp(tw,FOLDVAL);
		functions->DrawEmbossedProp(tw,FIXVAL);
	}
	if (refresh & 32) {
		functions->EnableEmbossed(tw,DIS_SYSEX,tool->action & REALTIME_SYSEX);
		functions->SelectEmbossed(tw,DO_SYSEX,tool->action & REALTIME_SYSEX);
		functions->SelectEmbossed(tw,DIS_SYSEX,tool->action & DISCARD_SYSEX);
	}
}


/*==========================================================================
* doevent
*
*  INPUT: pointer to event
* RETURN: pointer to event
*
* EFFECT: does something to an event
* 12 Nov 95: free noteoffs for branching; dupe noteons to branch
* 31 Mar 99: didn't work from the toolpad (PADEDIT not set!); added pad flag.
*=========================================================================*/

static struct Event *doevent(struct FiltTool *tool,struct Event *event,UBYTE pad)

{
	if (IsNoteOff(event)) { //12 Nov 95
		functions->freeevent(event);
		return(NULL);
	}
	switch (tool->action & ACTION_MASK) {
		case ACT_FOLD :
			if (tool->data & DATA_BYTE1) {
				if (!(tool->action & ACT_NOTCH)) {
					while (event->byte1 < tool->low) event->byte1 += tool->fold;
					while (event->byte1 > tool->high) event->byte1 -= tool->fold;
				}
				else { /* notch filter */
					if (event->byte1 <= tool->low && event->byte1 >= tool->high) {
						if (event->byte1 < (tool->low-tool->high>>1)+tool->high) { /* below midpoint */
							while (event->byte1 >= tool->high && event->byte1 < 255) 
								event->byte1 -= tool->fold; /* fold down */
							while (event->byte1 > 127) /* folded down past 0? */
								event->byte1 -= tool->fold;
						}
						else { /* above midpoint; fold up */
							while (event->byte1 <= tool->low && event->byte1 < 128)
								event->byte1 += tool->fold;
							while (event->byte1 >= 127) /* folded up too far? */
								event->byte1 += tool->fold;
						}
					}
				}
			}
			if (tool->data & DATA_BYTE2) {
				if (!(tool->action & ACT_NOTCH)) {
					while (event->byte2 < tool->low) event->byte2 += tool->fold;
					while (event->byte2 > tool->high) event->byte2 -= tool->fold;
				}
				else { /* notch filter */
					if (event->byte2 <= tool->low && event->byte2 >= tool->high) {
						if (event->byte2 <= (tool->low-tool->high>>1)+tool->high) { /* below midpoint */
							while (event->byte2 >= tool->high && event->byte2 < 255) 
								event->byte2 -= tool->fold; /* fold down */
							while (event->byte2 >= 127) /* folded down past 0? */
								event->byte2 -= tool->fold;
						}
						else { /* above midpoint; fold up */
							while (event->byte2 <= tool->low && event->byte2 < 128)
								event->byte2 += tool->fold;
							while (event->byte2 >= 127) /* folded up too far? */
								event->byte2 += tool->fold;
						}
					}
				}
			}
			break;

		case ACT_FIX :
			if (!(tool->action & ACT_NOTCH)) {
				if (tool->data & DATA_BYTE1) {
					if ((event->byte1 < tool->low)||(event->byte1 > tool->high)) {
						event->byte1=tool->fix;
					}
				}
				if (tool->data & DATA_BYTE2) {
					if ((event->byte2 < tool->low)||(event->byte2 > tool->high)) {
						event->byte2=tool->fix;
					}
				}
			}
			else { /* notch */
				if (tool->data & DATA_BYTE1) {
					if (event->byte1 >= tool->high && event->byte1 <= tool->low) {
						event->byte1=tool->fix;
					}
				}
				if (tool->data & DATA_BYTE2) {
					if (event->byte2 >= tool->high && event->byte2 <= tool->low) {
						event->byte2=tool->fix;
					}
				}
			}
			break;

		case ACT_BRANCH :
		case ACT_DISCARD :
			if (!(tool->action & ACT_NOTCH)) { /* bandpass */
				if ( ((tool->data & DATA_BYTE1) && 
					   ((event->byte1 < tool->low)||(event->byte1 > tool->high))) ||
						 (((tool->data & DATA_BYTE2) && !IsNoteOff(event)) &&
						 ((event->byte2 < tool->low)||(event->byte2 > tool->high))) ||
						 !((tool->data & DATA_BYTE1) | (tool->data & DATA_BYTE2)) ) {
					if ((tool->action & ACT_BRANCH) && tool->tool.branch) {
						event->tool=tool->tool.branch;
						event->type |= EVENT_BRANCH;
					}
					else { /* if discard or no branch, bag it */
						functions->freeevent(event);
						event=NULL;
					}
				}
			}
			else { /* notch */
				if ( ((tool->data & DATA_BYTE1) && 
				     ((event->byte1 <= tool->low)&&(event->byte1 >= tool->high))) ||
					   (((tool->data & DATA_BYTE2) && !IsNoteOff(event)) &&
					   ((event->byte2 <= tool->low)&&(event->byte2 >= tool->high))) ||
					   !((tool->data & DATA_BYTE1) | (tool->data & DATA_BYTE2)) ) {
					if ((tool->action & ACT_BRANCH) && tool->tool.branch) {
						event->tool=tool->tool.branch;
						event->type |= EVENT_BRANCH;
					}
					else { /* if discard or no branch, bag it */
						functions->freeevent(event);
						event=NULL;
					}
				}
			}
			break;

	} /* switch */
	
	if (event && !pad) {
dprintf("realtime! type=%ld\n",event->type);
			if (IsNoteOn(event)) {
//	if (event && IsNoteOn(event) && !(event->type & EVENT_PADEDIT)) { // 12 Nov 95
		if (event->next = functions->dupeevent(event)) {
				event->next->status = MIDI_NOTEOFF;
				event->next->time += ((struct NoteEvent *)event->next)->duration;
				event->next->byte2 = 0;
			}
		}
	}
	return(event);
}

/*==========================================================================
* processlist
*
*  INPUT: pointers to tool and list to process, start and stop times
* RETURN: pointer to processed list
*
* EFFECT: if appropriate, calls doevent() for each event in list
*=========================================================================*/

#define dothisnote(t,e) ( ((t)->event==0) && IsNote((e)) )
#define dothisctrl(t,e) ( IsControlMsg(e) && ((t)->event>0 && (t)->event<5) && ((e)->status==(t)->control) )

static struct Event * __saveds __stdargs processlist(struct FiltTool *tool,
                                                     struct Event *list,
														           long start, long stop)
{
	struct Event *event; 
	struct Event *prev; 
	struct Event *next;

	if (!tool->tool.touched) inittool(tool);

	prev = NULL;
	event = list; 
	while (event && (event->time < stop)) {
		next = event->next;
		event->tool = event->tool->next;
		if (dothisnote(tool,event) || dothisctrl(tool,event)) {
			event = doevent(tool,event,1);
			if (!event) {  /* freed event */
				if (prev) prev->next=next;
				else list = next;
dprintf("freed!?\n");
			}
		}
		if (event) {
			prev = event;
		}
		event = next;
	}
	return(list);
}


#define SX_LENGTH 9

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static struct Event * __inline dosysex(struct FiltTool *tool,
                                      struct StringEvent *event)

/*?? shouldn't there be a test or two on validity of data ??*/

{
	register unsigned char *sxdata;
	UBYTE refresh = 0;

	if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
		if (*sxdata != MIDI_EOX) {
			/* count how many bytes before eox */
			for ( ; (refresh < SX_LENGTH) && ((*(sxdata + refresh)) != MIDI_EOX) ;
			       refresh++ )
				;
			if (refresh) { /* if there are some, copy them */
				memcpy(&tool->action,sxdata,refresh);
				if (tool->tool.touched & TOUCH_REFRESHOK) {
					refreshgadgets(tool,255);  /* redraw all gadgetry */
				}
			}
		}

		if (tool->action & DISCARD_SYSEX) {
			functions->freestring(event->string);
			functions->freeevent((struct Event *)event);
			event=NULL;
		}
	}
	return((struct Event *)event);
}


/*==========================================================================
* processevent
*
*  INPUT: pointer to NoteEvent
* RETURN: pointer to Event
*
* EFFECT: checks some stuff and then calls doevent() if appropriate
*=========================================================================*/

static struct Event * __saveds __stdargs processevent(struct Event *event)

{
	struct FiltTool *tool = (struct FiltTool *)event->tool;

	if (!tool->tool.touched) inittool(tool);

	event->tool = event->tool->next;
/*! modify dothisnote() and put this in doevent() for sysex in the toolpad */
	if (IsSysEx(event) && (tool->action & REALTIME_SYSEX)) {
		event = dosysex(tool,(struct StringEvent *)event);
	}
	else if (dothisnote(tool,event) || dothisctrl(tool,event)) {
		event = doevent(tool,event,0);
	}
	return(event);
}

/* for display of a prop gadget value */
#define XOffset 46 /* relative to g->LeftEdge + g->Width (allowing 36 for clickers) */
#define YOffset 6  /* relative to g->TopEdge */
#define TEXTCOLOR 1

/*==========================================================================
* hilo1
*
*  INPUT: pointers to window and gadget, gadget value
* RETURN: gadget value
*
* EFFECT: display a pot value with a minimum of one
*=========================================================================*/

static long __saveds __stdargs hilo1(w,g,h)

struct Window *w;
struct Gadget *g;
long h;

{
	char text[4];

	if (h < 1) h=1;
	SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset,
	         (g->Flags & GADGDISABLED)?BP_DGREY:BP_WHITE,0,JAM2);
	sprintf(text,"%3ld",h);
	Text(w->RPort,text,3);
	return(h);
}

/*==========================================================================
* typehilo
*
*  INPUT: pointers to window and gadget, gadget value
* OUTPUT: none
* RETURN: gadget value
*
* EFFECT: display a pot value mapped to event type
*=========================================================================*/

#define XOff 42 /* relative to g->LeftEdge + g->Width*/
#define YOff 6  /* relative to g->TopEdge */

static long __saveds __stdargs typehilo(w,g,h)

struct Window *w;
struct Gadget *g;
long h;

{
	static char *eventtypes[] = {
		"Note On/Off   ",  /* type 0 */
		"Control Change",  /* type 1 */
		"Program Change",  /* type 2 */
		"Mono A-Touch  ",  /* type 3 */
		"Pitch Bend    "   /* type 4 */
	};

	SetRPort(w->RPort,g->LeftEdge+g->Width+XOff,g->TopEdge+YOff,TEXTCOLOR,0,JAM2);
	Text(w->RPort,eventtypes[h],14);
	return(h);
}


/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: makes a sysex message and posts it
*=========================================================================*/

static void __inline dosnapshot(struct FiltTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;
	
	memcpy(sx,&tool->action,SX_LENGTH);
	if (event = BuildSnapshot(&tool->tool, 
	                          functions->running ? functions->timenow : -1,
	                          sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
		if (tool->tool.track->edit->window)
			functions->updateobject(&functions->inittempo,0); /* redisplay */
	}
}

/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __saveds __stdargs edittool(struct FiltTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	UBYTE refresh = 255;

	if (!tool->tool.touched) inittool(tool);

	if (!(nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,20))) return;
	AddSnapshotGadget(tw);
	SetMenuStrip(tw,&Menu1);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	EmbossOnGList(tw,&GadgetList1,GCOUNT,1);
	functions->FatEmbossedPropOn(tw,EVENT,EVENTL,EVENTR,typehilo,5,1);
	functions->FatEmbossedPropOn(tw,CONTROLLER,CONTROLLERL,CONTROLLERR,hilo,128,1);
	functions->FatEmbossedPropOn(tw,HIGH,HIGHL,HIGHR,hilo,128,1);
	functions->FatEmbossedPropOn(tw,LOW,LOWL,LOWR,hilo,128,1);
	functions->FatEmbossedPropOn(tw,FOLDVAL,FOLDVALL,FOLDVALR,hilo1,37,1);
	functions->FatEmbossedPropOn(tw,FIXVAL,FIXVALL,FIXVALR,hilo,128,1);
	tool->tool.touched |= TOUCH_REFRESHOK;

	/*** ui handling ***/
	for ( ; ; ) {
		if (refresh) {
			refreshgadgets(tool,refresh);
			refresh=0;
		}
		message = functions->GetIntuiMessage(tw);
		code = message->Code;
		class = message->Class;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			tool->tool.touched &= ~TOUCH_REFRESHOK;
			break;
		}
		else if (class == GADGETUP) {
			switch (gadget->GadgetID) {
				case SNAPSHOT_GADGET_ID :
					dosnapshot(tool);
					break;
				case BYTE1 :
					tool->data ^= DATA_BYTE1;
					refresh=2;
					break;
				case BYTE2 :
					tool->data ^= DATA_BYTE2;
					refresh=2;
					break;
				case FOLD :
					tool->action = (tool->action & ACTION_HIBITS) | ACT_FOLD;
					refresh=20;
					break;
				case FIX :
					tool->action = (tool->action & ACTION_HIBITS) | ACT_FIX;
					refresh=20;
					break;
				case BRANCH :
					tool->action = (tool->action & ACTION_HIBITS) | ACT_BRANCH;
					refresh=20;
					break;
				case DISCARD :
					tool->action = (tool->action & ACTION_HIBITS) | ACT_DISCARD;
					refresh=20;
					break;
				case DO_SYSEX :
					tool->action ^= REALTIME_SYSEX;
					refresh = 32;
					break;
				case DIS_SYSEX :
					tool->action ^= DISCARD_SYSEX;
					refresh = 32;
					break;
			}
		}
		else if (class == GADGETDOWN) {
			switch (gadget->GadgetID) {
				case EVENT :
					tool->event = functions->DragEmbossedProp(tw,EVENT);
					refresh=1;
					break;
				case EVENTL :
					tool->event = functions->ShiftEmbossedProp(tw,EVENT,-1,0);
					refresh=1;
					break;
				case EVENTR :
					tool->event = functions->ShiftEmbossedProp(tw,EVENT,1,0);
					refresh=1;
					break;

				case CONTROLLER : 
					tool->controller = functions->DragEmbossedProp(tw,CONTROLLER);
					refresh=1;
					break;
				case CONTROLLERL :
					tool->controller = functions->ShiftEmbossedProp(tw,CONTROLLER,-1,0);
					refresh=1;
					break;
				case CONTROLLERR :
					tool->controller = functions->ShiftEmbossedProp(tw,CONTROLLER,1,0);
					refresh=1;
					break;

				/* code is a temporary repository to allow sanity checks */
				case FOLDVAL :
					code = functions->DragEmbossedProp(tw,FOLDVAL);
					break;
				case FOLDVALL :
					code = functions->ShiftEmbossedProp(tw,FOLDVAL,-1,0);
					break;
				case FOLDVALR :
					code = functions->ShiftEmbossedProp(tw,FOLDVAL,1,0);
					break;

				case FIXVAL : /* no sanity required */
					tool->fix = functions->DragEmbossedProp(tw,FIXVAL);
					refresh=16;
					break;
				case FIXVALL :
					tool->fix = functions->ShiftEmbossedProp(tw,FIXVAL,-1,0);
					refresh=16;
					break;
				case FIXVALR :
					tool->fix = functions->ShiftEmbossedProp(tw,FIXVAL,1,0);
					refresh=16;
					break;

				case LOW :
					code = functions->DragEmbossedProp(tw,LOW);
					break;
				case LOWL :
					code = functions->ShiftEmbossedProp(tw,LOW,-1,0);
					break;
				case LOWR :
					code = functions->ShiftEmbossedProp(tw,LOW,1,0);
					break;

				case HIGH :
					code = functions->DragEmbossedProp(tw,HIGH);
					break;
				case HIGHL :
					code = functions->ShiftEmbossedProp(tw,HIGH,-1,0);
					break;
				case HIGHR :
					code = functions->ShiftEmbossedProp(tw,HIGH,1,0);
					break;
			}
			/* do some sanity checking on the new pot values */
			switch (gadget->GadgetID) {
				case HIGH  :
				case HIGHL :
				case HIGHR :
					if (code <= tool->low) tool->action |= ACT_NOTCH;
					else tool->action &= 0x7F;

					if (tool->action & ACT_FOLD && !(tool->action & ACT_NOTCH)) { /* spread can't be less than fold */
						if (code >= tool->low+tool->fold) tool->high = code;
						else if (code < tool->low+tool->fold) tool->high = tool->low+tool->fold;
					}
					else if (tool->action & ACT_FOLD && tool->action & ACT_NOTCH) {
						if (code+(127-tool->low) >= tool->fold) tool->high = code;
						else tool->high = tool->fold - (127-tool->low);
					}
					else if (!(tool->action & ACT_NOTCH)) {
						tool->high = (code) ? code : 1 ;
						if (tool->high == tool->low) {
							if (tool->low < 127) tool->high = tool->low+1;
							else tool->low = tool->high-1;
						}
					}
					else tool->high = code;
					refresh=24;
					break;

				case LOW  :
				case LOWL :
				case LOWR :
					if (code >= tool->high) tool->action |= ACT_NOTCH;
					else tool->action &= 0x7F;

					if (tool->action & ACT_FOLD && !(tool->action & ACT_NOTCH)) { /* spread can't be less than fold */
						if (code <= tool->high-tool->fold) tool->low = code;
						else if (code > tool->high - tool->fold) tool->low = tool->high-tool->fold;
					}
					else if (tool->action & ACT_FOLD && tool->action & ACT_NOTCH) {
						if (tool->high+(127-code) >= tool->fold) tool->low = code;
						else tool->low = 127 - (tool->fold-tool->high);
					}
					else if (!(tool->action & ACT_NOTCH)) {
						tool->low = code;
						if (tool->low == tool->high) {
							if (tool->high) tool->low--;
							else tool->high++;
						}
					}
					else tool->low = code;
					refresh=24;
					break;

				case FOLDVAL  :
				case FOLDVALL :
				case FOLDVALR :
					if (tool->action & ACT_FOLD) {
						if (!(tool->action & ACT_NOTCH)) {
							if (code <= tool->high-tool->low) tool->fold = code;
							else if (code > tool->high) tool->fold = tool->high - 1;
							else if (code < tool->low) tool->fold = tool->low + 1;
						}
						else {
							if (code > tool->high + (127-tool->low)) {
								tool->fold = tool->high + (127-tool->low);
							}
							else tool->fold = code;
						}
						refresh=16;
					}
					break;
			}
#if DBUG
if (refresh) {
	dprintf("3: event=%ld controller=%ld action=%ld low=%ld high=%ld fold=%ld\n\n",tool->event,tool->controller,tool->action,tool->low,tool->high,tool->fold);
}
#endif
		}
	}

	EmbossOffGList(tw,&GadgetList1,GCOUNT);
	functions->FatEmbossedPropOff(tw,EVENT,EVENTL,EVENTR);
	functions->FatEmbossedPropOff(tw,CONTROLLER,CONTROLLERL,CONTROLLERR);
	functions->FatEmbossedPropOff(tw,LOW,LOWL,LOWR);
	functions->FatEmbossedPropOff(tw,HIGH,HIGHL,HIGHR);
	functions->FatEmbossedPropOff(tw,FOLDVAL,FOLDVALL,FOLDVALR);
	functions->FatEmbossedPropOff(tw,FIXVAL,FIXVALL,FIXVALR);
	NewCloseEmbossedWindow(&tool->tool,nw);
}


/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and initializes a tool with data read from disk.  This
*         is necessary to clear TOUCH_REFRESHOK in tool->tool.touched in case
*         the tool was saved while the edit window was open.
*=========================================================================*/

static struct Tool * __saveds __stdargs loadtool(long file, long size)

{
	struct FiltTool *tool;

	if (tool = (struct FiltTool *)functions->myalloc(TOOL_SIZE,MEMF_CLEAR)) {
		functions->fastread(file,(char *)tool,size);
		tool->tool.touched &= ~TOUCH_REFRESHOK;
	}
	return((struct Tool *)tool);
}



static struct ToolMaster master;

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to tool master struct
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __saveds __stdargs inittoolmaster()

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = THIS_TOOL_ID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.image = &FiltIconDown;
	master.upimage = &FiltIconUp;
	master.loadtool = loadtool;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL | TOOL_BRANCHOUT;

	return(&master);
}
